/*
  Macro           CursorShape
  Author          Carlo Hogeveen
  Website         eCarlo.nl/tse
  Compatibility   Linux TSE, probably far back compatible
  Version         v1.0.2   17 Feb 2024


  This extension allows us to set separate Insert and Overwrite cursor shapes.

  BUT BUT BUT
  It does not work for all Linux brands in combination with all command line
  interfaces or all terminals.

  For example, for me the CursorShape extension works
    in Debian within WSL, and
    directly acessing a (virtual) Debian desktop machine,
  but not
    directly acessing a (virtual) Debian server, or
    indirectly accesing anything through Putty.

  You will have to test yourself for which, if any, Linux brand/CLI/terminal
  combination the CursorShape extension works for you.

  My hope is, that it will at least work on all WSL Linux installations and on
  all modern Linux desktops.


  INSTALLATION & CONFIGURATION

  Copy this file to TSE's "mac" directory and compile it there.

  The only configuration action needed is to execute this extension as a macro
  and selecting an Insert and an Overwrite cursor shape.

  The extension cannot set the shape for only one: If you do not set both, then
  the extension will set the other one to the same value (initially invisibly).


  INFO

  I do not know, what the difference between "Blinking Block" and "Blinking
  Default Block" is. They do not make a difference in my environments, but
  maybe they do in others, so they are both there to try.


  HISTORY

  v1       16 Feb 2024
    Initial release.

  v1.0.1   17 Feb 2024
    Added a proper warning if it cannot save its configuration.

  v1.0.2   17 Feb 2024
    As suggested by Sammy, I changed the Dos() call to an fWrite(_STDOUT_) call.
    This is technically a better solution because it does not need to start
    another process.

*/





/*
  T E C H N O L O G I C A L   B A C K G R O U N D   I N F O R M A T I O N

  The ANSI escape codes to set the cursor shape are based on the web page below.
  Search it for "cursor style".

  https://invisible-island.net/xterm/ctlseqs/ctlseqs.html#h2-Functions-using-CSI-_-ordered-by-the-final-character_s_

*/





// Start of compatibility restrictions and mitigations


#ifndef LINUX

  This extension is only for LINUX, not for Windows.

#endif


// End of compatibility restrictions and mitigations.





// Global constants and semi constants.
string MACRO_NAME [MAXSTRINGLEN] = ''


// Global variables.
string  cur_shape_id       [1] = ''
integer input_mode             = -1
string  insert_shape_id    [1] = ''
string  overwrite_shape_id [1] = ''
integer recursion              = 0
integer skip_configuration     = FALSE


integer proc write_profile_str(string section_name,
                               string item_name,
                               string item_value)
  integer result = WriteProfileStr(section_name,
                                   item_name,
                                   item_value)
  if not result
    Warn(MACRO_NAME; 'error: Cannot save configuration to "tse.ini".')
  endif
  return(result)
end write_profile_str


proc idle()
  string new_shape_id [1] = ''

  if Query(Insert) <> input_mode
    input_mode   = Query(Insert)
    new_shape_id = iif(input_mode, insert_shape_id, overwrite_shape_id)

    if new_shape_id <> cur_shape_id
      cur_shape_id = new_shape_id
      fWrite(_STDOUT_, Chr(27) + '[' + cur_shape_id + ' q')
      UpdateDisplay()

      // Allow configuration from Main() if it was disallowed.
      skip_configuration = FALSE
    endif
  endif
end idle


proc on_abandon_editor()
  fWrite(_STDOUT_, Chr(27) + 'c')
end on_abandon_editor


proc reset_shapes()
  RemoveProfileSection(MACRO_NAME + ':Config')
  DelAutoLoadMacro(MACRO_NAME)
end reset_shapes


string proc get_shape_name(string input_mode)
  string shape_id    [2] = ''
  string shape_name [22] = ''

  if input_mode == 'Insert'
    shape_id = insert_shape_id
  else
    shape_id = overwrite_shape_id
  endif

  case shape_id
    when '0'
      shape_name = 'Blinking block'
    when '1'
      shape_name = 'Blinking default block'
    when '2'
      shape_name = 'Steady block'
    when '3'
      shape_name = 'Blinking underline'
    when '4'
      shape_name = 'Steady underline'
    when '5'
      shape_name = 'Blinking bar'
    when '6'
      shape_name = 'Steady bar'
  endcase

  return(shape_name)
end get_shape_name


proc set_shape(string input_mode_name, string new_shape_id)
  string other_mode [9] = iif(input_mode_name == 'Insert', 'Overwrite', 'Insert')

  // Extra precaution to avoid infinite recursion if a WriteProfileStr() failed
  // or if tse.ini is corrupt or manipulated.
  recursion = recursion + 1

  if  recursion <= 2
  and (new_shape_id in '0' .. '6')
  and write_profile_str(MACRO_NAME + ':Config', input_mode_name + 'Shape', new_shape_id)
    if input_mode_name == 'Insert'
      insert_shape_id    = new_shape_id
    else
      overwrite_shape_id = new_shape_id
    endif

    input_mode = -1 // Triggers changing the cursor in idle().

    if not isAutoLoaded()
      AddAutoLoadMacro(MACRO_NAME)
    endif

    // Cannot explicitly only set the external cursor, only reset everything.
    // Therefore both input modes need an internal shape (or neither).
    if     other_mode         == 'Insert'
    and    insert_shape_id    == ''
      set_shape('Insert'   , new_shape_id)
    elseif other_mode         == 'Overwrite'
    and    overwrite_shape_id == ''
      set_shape('Overwrite', new_shape_id)
    endif
  endif

  recursion = recursion - 1
end set_shape


menu configure_overwrite_shape()
  title   = 'Configure cursor Overwrite shape'
  history = iif((overwrite_shape_id in '0' .. '6'), Val(overwrite_shape_id) + 1, 1)

  "Blinking block",
    set_shape('Overwrite', '0'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Blinking default block",
    set_shape('Overwrite', '1'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Steady block",
    set_shape('Overwrite', '2'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Blinking underline",
    set_shape('Overwrite', '3'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Steady underline",
    set_shape('Overwrite', '4'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Blinking bar",
    set_shape('Overwrite', '5'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Steady bar",
    set_shape('Overwrite', '6'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'
end configure_overwrite_shape


menu configure_insert_shape()
  title   = 'Configure cursor Insert shape'
  history = iif((insert_shape_id in '0' .. '6'), Val(insert_shape_id) + 1, 1)

  "Blinking block",
    set_shape('Insert', '0'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Blinking default block",
    set_shape('Insert', '1'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Steady block",
    set_shape('Insert', '2'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Blinking underline",
    set_shape('Insert', '3'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Steady underline",
    set_shape('Insert', '4'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Blinking bar",
    set_shape('Insert', '5'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'

  "Steady bar",
    set_shape('Insert', '6'),
    _MF_CLOSE_AFTER_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'
end configure_insert_shape


menu configuration_menu()
  title   = 'Configure TSE cursor shapes in Linux.'
  history

  "Cursor shape: &Insert"
    [get_shape_name('Insert'):22],
    configure_insert_shape(),
    _MF_DONT_CLOSE_|_MF_ENABLED_,
    'Changing cursor shapes does not work for every Linux/CLI/Terminal combination.'
  "Cursor shape: &Overwrite"
    [get_shape_name('Overwrite'):22],
    configure_overwrite_shape(),
    _MF_DONT_CLOSE_|_MF_ENABLED_,
    'Changing a cursor shape does not work for every Linux/CLI/Terminal combination.'
  "&Reset both to default cursor when TSE restarts",
    reset_shapes(),,
    'Reset both cursur shapes to the external cursor after TSE restarts.'
end configuration_menu


proc WhenPurged()
end WhenPurged


proc WhenLoaded()
  MACRO_NAME         = SplitPath(CurrMacroFilename(), _NAME_)
  insert_shape_id    = GetProfileStr(MACRO_NAME + ':Config', 'InsertShape'   , '')
  overwrite_shape_id = GetProfileStr(MACRO_NAME + ':Config', 'OverwriteShape', '')

  if not (insert_shape_id    in '0' .. '6')
  or not (overwrite_shape_id in '0' .. '6')
    configuration_menu()
    skip_configuration = TRUE
  endif

  Hook(_IDLE_             , idle             )
  Hook(_ON_ABANDON_EDITOR_, on_abandon_editor)
end WhenLoaded


proc Main()
  if skip_configuration
    skip_configuration = FALSE
  else
    configuration_menu()
  endif
end Main

